/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 10/12/22.
//

#ifndef FGGL_GRID_HEXAGON_BOARD_HPP
#define FGGL_GRID_HEXAGON_BOARD_HPP

#include <map>
#include <set>
#include <optional>

#include "fggl/grid/hexagon.hpp"
#include "fggl/math/types.hpp"

namespace fggl::grid {

	struct MaterialData {
		std::string name;
		math::vec3 colour;
	};

	struct TerrainType {
		std::shared_ptr<MaterialData> data;
	};

	struct HexTile {
		std::shared_ptr<MaterialData> terrain;

		[[nodiscard]]
		inline std::optional<const MaterialData> data() const {
			if (terrain == nullptr) {
				{}
			}

			return *terrain.get();
		}
	};

	class HexGrid {
		public:
			inline bool isValidPos(const IntHex& pos) const {
				return m_tiles.contains(pos);
			}

			void setTerrain(const IntHex& pos, const TerrainType& terrain) {
				auto& mapTile = m_tiles[pos];
				mapTile.terrain = terrain.data;
			}

			std::optional<const MaterialData> getTerrain(const IntHex& pos) const {
				const auto itr = m_tiles.find(pos);
				if ( itr == m_tiles.end() ) {
					return {};
				}
				return itr->second.data();
			}

			std::set<IntHex> getAllTiles() {
				std::set<IntHex> posSet;
				for ( auto& [pos,data] : m_tiles ) {
					posSet.emplace( pos );
				}
				return posSet;
			}

			std::set<HexTile> tilesInRange(const IntHex& pos, int range) const;
			std::set<HexTile> neighboursOf(const IntHex& pos) const;

		private:
			std::map<IntHex, HexTile> m_tiles;
	};

} // namespace fggl::grid

#endif //FGGL_GRID_HEXAGON_BOARD_HPP
