/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

/*
 * This file is part of FGGL.
 *
 * FGGL is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any
 * later version.
 *
 * FGGL is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with FGGL.
 * If not, see <https://www.gnu.org/licenses/>.
 */

//
// Created by webpigeon on 18/12/22.
//

#ifndef FGGL_FGGL_GRID_LAYOUT_HPP
#define FGGL_FGGL_GRID_LAYOUT_HPP

#include "fggl/math/types.hpp"
#include "fggl/grid/hexagon.hpp"

namespace fggl::grid {

	const math::mat2 MAT_HEX_POINTY {
		std::sqrt(3.0F), 0.0F,
		std::sqrt(3.0F) / 2.0F, 3.0F/2.0F
	};

	const math::mat2 MAT_HEX_FLAT {
		3.F/2.F, std::sqrt(3.F)/2.F,
		0.0F, std::sqrt(3.F)
	};

	struct Orientation {
		math::mat2 m_forward;
		math::mat2 m_backward;
		int m_angle;

		inline Orientation(int angle, math::mat2 forward) : m_forward(forward), m_backward(glm::inverse(forward)), m_angle(angle) {};

		static inline Orientation make_pointy() {
			return { 30, MAT_HEX_POINTY};
		}

		static inline Orientation make_flat() {
			return { 0, MAT_HEX_FLAT };
		}
	};


	struct Layout {
		Orientation m_orientation;
		math::vec2 m_size;
		math::vec2 m_origin;

		Layout(Orientation orientation, math::vec2 size, math::vec2 origin) : m_orientation(orientation), m_size(size), m_origin(origin){}
		Layout(Orientation orientation, float size) : m_orientation(orientation), m_size(size, size), m_origin() {}

		[[nodiscard]]
		inline math::vec2 origin() const {
			return m_origin;
		}

		[[nodiscard]]
		inline math::vec2 size() const {
			return m_size;
		}

		[[nodiscard]]
		inline math::vec2 toScreen(IntHex gridPos) const {
			math::vec2 hexPoint{gridPos.q(), gridPos.r()};
			auto p = (m_orientation.m_forward * hexPoint) * m_size;
			return p + m_origin;
		}

		[[nodiscard]]
		inline FloatHex toGrid(math::vec2 screen) const {
			auto point = (screen - m_origin) / m_size;
			auto p = m_orientation.m_backward * point;
			return {p.x, p.y};
		}

		[[nodiscard]]
		inline math::vec2 cornerOffset(int corner) const {
			int angInc = (360 / 6) * corner;
			float angle = (angInc + m_orientation.m_angle) * ( fggl::math::PI / 180.0F);
			return {
				m_size.x * cosf(angle),
				m_size.y * sinf(angle)
			};
		}

		void paintHex(fggl::gfx::Paint& paint, IntHex pos, math::vec3 colour) const {
			const auto hexScreenCenter = toScreen(pos);

			gfx::Path2D path({0,0});
			path.colour(colour);

			for (int i=0; i < 6; ++i) {
				auto cornerPos =  hexScreenCenter + cornerOffset(i);
				if ( i == 0) {
					path.moveTo(cornerPos);
				} else {
					path.pathTo(cornerPos);
				}
			}
			paint.stroke(path);
		}


	};

}

#endif //FGGL_FGGL_GRID_LAYOUT_HPP
